﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using Windows.Media.Audio;
using Windows.Media.MediaProperties;
using Windows.Media;
using WinRT;

namespace MediaPlayer_vs_AudioGraph_Demo
{

    [ComImport]
    [Guid("5B0D3235-4DBA-4D44-865E-8F1D0E4FD04D")]
    [InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
    unsafe interface IMemoryBufferByteAccess
    {
        void GetBuffer(out byte* buffer, out uint capacity);
    }

    class WaveformStreamWriter : WaveformBuilder
    {
        readonly Stream stream;

        public WaveformStreamWriter(Stream stream)
        {
            this.stream = stream;
        }

        protected override void WriteBytes(byte[] bytes)
        {
            stream.Write(bytes, 0, bytes.Length);
        }
    }

    internal class AudioGraphSignalOutput : ISignalOutput
    {
        private AudioGraph audioGraph;
        private AudioDeviceOutputNode audioDeviceOutputNode;
        private AudioFrameInputNode audioFrameInputNode;

        private byte[] frameBytes;
        private AudioFrame leadFrame;
        private AudioFrame repeatFrame;
        private bool leadAFplayed;

        public bool RepeatFrames
        {
            get;
            set;
        }

        public bool Repeat1stFrame
        {
            get;
            set;
        }

        public AudioGraphSignalOutput()
        {
            Initialize();
        }

        ~AudioGraphSignalOutput()
        {
            if (audioGraph != null)
            {
                audioGraph.Dispose();
            }
        }

        private async Task Initialize()
        {
            var audioGraphSettings = new AudioGraphSettings(Windows.Media.Render.AudioRenderCategory.Media);
            audioGraphSettings.EncodingProperties =
                AudioEncodingProperties.CreatePcm(CarrierSampleSet.SampleRate, 1, 8);
            var audioGraphCreationResult = await AudioGraph.CreateAsync(audioGraphSettings);
            if (audioGraphCreationResult.Status != AudioGraphCreationStatus.Success)
            {
                throw audioGraphCreationResult.ExtendedError;
            }

            audioGraph = audioGraphCreationResult.Graph;

            var deviceOutputNodeCreationResult = await audioGraph.CreateDeviceOutputNodeAsync();
            if (deviceOutputNodeCreationResult.Status != AudioDeviceNodeCreationStatus.Success)
            {
                throw deviceOutputNodeCreationResult.ExtendedError;
            }

            audioDeviceOutputNode = deviceOutputNodeCreationResult.DeviceOutputNode;

            var encProps = AudioEncodingProperties.CreatePcm(CarrierSampleSet.SampleRate, 1, 8);
            audioFrameInputNode = audioGraph.CreateFrameInputNode(encProps);
            audioFrameInputNode.AddOutgoingConnection(audioDeviceOutputNode);
            audioFrameInputNode.Stop(); // Initialize the Frame Input Node in the stopped state
            audioFrameInputNode.QuantumStarted += AudioFrameInputNode_QuantumStarted;
            audioFrameInputNode.AudioFrameCompleted += AudioFrameInputNode_AudioFrameCompleted;

            //audioGraph.Start(); // Start the graph since we will only start/stop the frame input node
        }

        private void AudioFrameInputNode_AudioFrameCompleted(AudioFrameInputNode sender, AudioFrameCompletedEventArgs args)
        {
        }

        private void AudioFrameInputNode_QuantumStarted(AudioFrameInputNode sender, FrameInputNodeQuantumStartedEventArgs args)
        {
            if (args.RequiredSamples == 0)
            {
                return;
            }

            // TODO Maybe feed exactly needed count of bytes? - Didn't work out

            if (RepeatFrames && Repeat1stFrame)
            {
                sender.AddFrame(leadFrame);
            }
            else if (!leadAFplayed)    // It's time to play the lead frame
            {
                sender.AddFrame(leadFrame);
                leadAFplayed = true;
            }
            else if (RepeatFrames)  // It's time to play the repeat frame
            {
                if (repeatFrame == null)   // We didn't play it before, so generate it first
                {
                    repeatFrame = CreateFrame(frameBytes[0], frameBytes[1], frameBytes[2], repeatFrame: true);
                }
                sender.AddFrame(repeatFrame);
            }
            else
            {
                sender.Stop();
                audioGraph.Stop();
                leadAFplayed = false;
            }
        }

        public void StartTransmit(byte lowCustom, byte highCustom, byte data)
        {
            // If frames are already generated and no changes - don't regenerate
            if (leadFrame != null && frameBytes != null &&
                frameBytes[0] == lowCustom && frameBytes[1] == highCustom && frameBytes[2] == data)
            {
                ;
            }
            else
            {
                if (leadFrame != null)
                    leadFrame.Dispose();
                leadFrame = CreateFrame(lowCustom, highCustom, data);
                frameBytes = new byte[3] { lowCustom, highCustom, data };
            }

            audioGraph.Start();
            audioFrameInputNode.Start();
        }

        private unsafe AudioFrame CreateFrame(byte lowCustom, byte highCustom, byte data, bool repeatFrame = false)
        {
            MemoryStream memoryStream = new MemoryStream();
            var waveformBuilder = new WaveformStreamWriter(memoryStream);
            if (repeatFrame)
            {
                waveformBuilder.WritePauseCode(25);    // To not repeat too fast (maybe it's better to pause somehow else?)
                waveformBuilder.WriteRepeatCode();
            }
            else
            {
                waveformBuilder.WritePauseCode(250);    // To not repeat too fast (maybe it's better to pause somehow else?)
                waveformBuilder.WriteLeadCode();
            }
            waveformBuilder.WriteByteCode(lowCustom);
            waveformBuilder.WriteByteCode(highCustom);
            waveformBuilder.WriteByteCode(data);
            waveformBuilder.WriteByteCode((byte)~data);
            waveformBuilder.WriteBitCode(false);    // Critical: Final pulse to mark end of the last bit
            memoryStream.Flush();

            AudioFrame frame = new AudioFrame((uint)memoryStream.Length);
            using var buffer = frame.LockBuffer(AudioBufferAccessMode.Write);
            using var reference = buffer.CreateReference();
            reference.As<IMemoryBufferByteAccess>().GetBuffer(out byte* pBuffer, out uint pBufferCapacity);

            UnmanagedMemoryStream unMemoryStream = new UnmanagedMemoryStream(
                pBuffer, pBufferCapacity, pBufferCapacity, FileAccess.Write);

            //frame.IsDiscontinuous = true;
            memoryStream.WriteTo(unMemoryStream);   // Same as '.Position = 0; .CopyTo()'
            return frame;
        }
    }

}
