﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.Media.Core;
using Windows.Media.MediaProperties;
using Windows.Media.Playback;
using Windows.Storage.Streams;

namespace MediaPlayer_vs_AudioGraph_Demo
{

    class WaveformDataWriter : WaveformBuilder
    {
        readonly DataWriter dataWriter;

        public WaveformDataWriter(DataWriter dataWriter)
        {
            this.dataWriter = dataWriter;
        }

        protected override void WriteBytes(byte[] bytes)
        {
            dataWriter.WriteBytes(bytes);
        }
    }

    internal class MediaStreamSignalOutput : ISignalOutput
    {
        private readonly MediaPlayer mediaPlayer = new();
        private InMemoryRandomAccessStream memStream;
        private byte[] frameBytes;
        private ulong firstFrameSamplesCount;
        private ulong leadFramesBuffred;
        private ulong repeatFramesBuffred;

        public bool RepeatFrames
        {
            get;
            set;
        }

        public bool Repeat1stFrame
        {
            get;
            set;
        }

        public MediaStreamSignalOutput()
        {
            var encProps = AudioEncodingProperties.CreatePcm(CarrierSampleSet.SampleRate, 1, 8);
            var streamDesc = new AudioStreamDescriptor(encProps);
            var streamSource = new MediaStreamSource(streamDesc);
            streamSource.Starting += StreamSource_Starting;
            streamSource.SampleRequested += StreamSource_SampleRequested;
            streamSource.Closed += StreamSource_Closed;
            //streamSource.IsLive = true; // No effect?
            streamSource.BufferTime = TimeSpan.Zero;

            // if don't want to customize SMTC (System Media Transport Controls)
            //mediaPlayer.SetMediaSource(streamSource);

            // if want to customize SMTC (System Media Transport Controls)
            var mediaSource = MediaSource.CreateFromMediaStreamSource(streamSource);
            var playbackItem = new MediaPlaybackItem(mediaSource);
            var mediaItemDisplayProperties = playbackItem.GetDisplayProperties();
            mediaItemDisplayProperties.Type = Windows.Media.MediaPlaybackType.Music;
            mediaItemDisplayProperties.MusicProperties.Title = "LG Flatron Mx94D-PZ IR Control";
            mediaItemDisplayProperties.MusicProperties.Artist = "Media Stream Signal Output";
            playbackItem.ApplyDisplayProperties(mediaItemDisplayProperties);
            mediaPlayer.Source = playbackItem;

            mediaPlayer.RealTimePlayback = true;
        }

        private void StreamSource_Starting(MediaStreamSource sender, MediaStreamSourceStartingEventArgs args)
        {
            // Stream was prepared earlier, just set start to very beginning
            args.Request.SetActualStartPosition(TimeSpan.Zero);
            leadFramesBuffred = 0;
            repeatFramesBuffred = 0;
        }

        private async void StreamSource_SampleRequested(MediaStreamSource sender, MediaStreamSourceSampleRequestedEventArgs args)
        {
            var deferral = args.Request.GetDeferral();
            if (RepeatFrames && Repeat1stFrame)
            {
                memStream.Seek(0);
            }
            if (memStream.Position == 0)    // It's time to play the lead frame
            {
                args.Request.Sample = await MediaStreamSample.CreateFromStreamAsync(
                    memStream,
                    (uint)firstFrameSamplesCount,
                    leadFramesBuffred * TimeSpan.FromSeconds((double)firstFrameSamplesCount / CarrierSampleSet.SampleRate));
                args.Request.Sample.Duration = TimeSpan.FromSeconds((double)firstFrameSamplesCount / CarrierSampleSet.SampleRate);
                args.Request.Sample.KeyFrame = true;
                leadFramesBuffred++;
                //Debug.Print($"Buffered Lead frames: {leadFramesBuffred}");
            }
            else if (RepeatFrames)  // It's time to play the repeat frame
            {
                if (memStream.Position == firstFrameSamplesCount)   // We didn't play it before, so generate it first
                {
                    await WriteFrame(memStream, frameBytes[0], frameBytes[1], frameBytes[2], repeatFrame: true);
                }
                memStream.Seek(firstFrameSamplesCount);
                args.Request.Sample = await MediaStreamSample.CreateFromStreamAsync(
                    memStream,
                    (uint)(memStream.Size - firstFrameSamplesCount),
                    repeatFramesBuffred * TimeSpan.FromSeconds((double)(memStream.Size - firstFrameSamplesCount) / CarrierSampleSet.SampleRate)
                    + leadFramesBuffred * TimeSpan.FromSeconds((double)firstFrameSamplesCount / CarrierSampleSet.SampleRate));
                args.Request.Sample.Duration = TimeSpan.FromSeconds((double)(memStream.Size - firstFrameSamplesCount) / CarrierSampleSet.SampleRate);
                args.Request.Sample.KeyFrame = true;
                repeatFramesBuffred++;
                //Debug.Print($"Buffered Repeat frames: {repeatFramesBuffred}");
            }
            deferral.Complete();
        }

        private void StreamSource_Closed(MediaStreamSource sender, MediaStreamSourceClosedEventArgs args)
        {
            sender.Starting -= StreamSource_Starting;
            sender.SampleRequested -= StreamSource_SampleRequested;
            sender.Closed -= StreamSource_Closed;
        }

        public async void StartTransmit(byte lowCustom, byte highCustom, byte data)
        {
            // If stream is already generated and no changes - don't regenerate, just seek to beginning
            if (memStream != null && frameBytes != null &&
                frameBytes[0] == lowCustom && frameBytes[1] == highCustom && frameBytes[2] == data)
            {
                memStream.Seek(0UL);
            }
            else
            {
                if (memStream != null)
                    memStream.Dispose();
                memStream = new InMemoryRandomAccessStream();
                await WriteFrame(memStream.GetOutputStreamAt(0), lowCustom, highCustom, data);
                firstFrameSamplesCount = memStream.Size;
                frameBytes = new byte[3] { lowCustom, highCustom, data };
            }

            mediaPlayer.Play();
        }

        async Task WriteFrame(IOutputStream stream, byte lowCustom, byte highCustom, byte data, bool repeatFrame = false)
        {
            using var dataWriter = new DataWriter(stream);
            var waveformBuilder = new WaveformDataWriter(dataWriter);
            if (repeatFrame)
            {
                waveformBuilder.WritePauseCode(25);    // To not repeat too fast (maybe it's better to pause somehow else?)
                waveformBuilder.WriteRepeatCode();
            }
            else
            {
                waveformBuilder.WritePauseCode(250);    // To not repeat too fast (maybe it's better to pause somehow else?)
                waveformBuilder.WriteLeadCode();
            }
            waveformBuilder.WriteByteCode(lowCustom);
            waveformBuilder.WriteByteCode(highCustom);
            waveformBuilder.WriteByteCode(data);
            waveformBuilder.WriteByteCode((byte)~data);
            waveformBuilder.WriteBitCode(false);    // Critical: Final pulse to mark end of the last bit
            await dataWriter.StoreAsync();
            await dataWriter.FlushAsync();
            dataWriter.DetachStream();  // To not close memStream during disposing DataWriter
        }
    }

}
